// $Id: pic.cc 269 2012-11-25 03:45:41Z peter $

// Copyright (C) 2005 Peter Johansson
// Copyright (C) 2006 Markus Ringnér, Peter Johansson
// Copyright (C) 2011, 2012 Peter Johansson
//
// This file is part of speclust, http://bioinfo.thep.lu.se/speclust.html
//
// Speclust is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 3 of the License, or (at your option) any later
// version.
//
// Speclust is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// speclust. If not, see <http://www.gnu.org/licenses/>.

#include "pic.h"

#include "Cluster.h"
#include "Parameter.h"
#include "Peak.h"
#include "Spectrum.h"

#include "DistancePearson.h"

#include <yat/utility/CommandLine.h>
#include <yat/utility/Matrix.h>
#include <yat/utility/OptionArg.h>
#include <yat/utility/OptionHelp.h>
#include <yat/utility/OptionFile.h>

#include <algorithm>
#include <fstream>
#include <iostream>
#include <iterator>
#include <sstream>
#include <vector>

int main(int argc, char* argv[])
{
  using namespace theplu::speclust;
  using namespace std;

	Parameter option(argc, argv);
  vector<vector<double> > masses;
  if(option.peakset_file().size()) {
		ifstream is(option.peakset_file().c_str());
		theplu::yat::utility::load(is, masses, '\0' , '\n', true, false);
		is.close();
	}
	else
		theplu::yat::utility::load(std::cin, masses, '\0' , '\n', true, false);

  std::vector<std::string> id;
  read_id(id, option.id_file().c_str());

  // Create Spectra and Peaks
  std::vector<Spectrum*> spectra;
  for (size_t i=0; i<masses.size(); i++) {
    Spectrum* s = new Spectrum(id[i]);
    for (size_t j=0; j<masses[i].size(); j++) {
      Peak* p = new Peak(s, masses[i][j]);
      s->push_back(p);
    }
    spectra.push_back(s);
  }
  // sort spectra in name order (alphabetically).
  std::sort(spectra.begin(), spectra.end(), NameOrder());

  // pairwise must be done before consensus and peaksets
  pairwise(spectra,option);
  if (option.consensus_file().size()){
    consensus(spectra, option);
  }
  if (option.multiple_file().size()){
    multiple(spectra, option);
  }

  for (size_t i=0; i<spectra.size(); i++){
    for (size_t j=0; j<spectra[i]->nof_peaks(); j++){
      delete spectra[i]->peaks()[j];
    }
    delete spectra[i];
  }

  exit(0);				// normal exit
}

namespace theplu {
namespace speclust {

  void pairwise(std::vector<Spectrum*>& spectra, const Parameter& option)
  {
    std::ofstream os(option.pairwise_file().c_str());
    os << "Width: " << option.sigma() << "\n"
			 << "Pairwise cutoff: " << option.pairwise_cutoff() << "\n\n";

    // Peter, could use any distance, but wanna keep base class abstract.
    DistancePearson distance;

    for (size_t i=0; i<spectra.size()-1; i++)
      for (size_t j=i+1; j<spectra.size(); j++) {
				yat::utility::Matrix dot_m;
				std::vector<std::pair<size_t,size_t> > path;
				distance.score(spectra[i]->mass(), spectra[j]->mass(),
											 option.sigma(), dot_m, path);

				// Check if any match is larger than cutoff
				bool write=false;
				for (size_t k=0; k<path.size(); k++)
					if (dot_m(path[k].first, path[k].second) > option.pairwise_cutoff()){
						write = true;
						break;
					}
				if (write) {
					// we want the peaks in ascending order (by mass).
					reverse(path.begin(),path.end());

					// first peakset
					os << spectra[i]->name();
					for (size_t k=0; k<path.size(); k++) {
						double match = dot_m(path[k].first, path[k].second);
						if (match > option.pairwise_cutoff()){
							Peak* peak1 = spectra[i]->peaks()[path[k].first];
							Peak* peak2 = spectra[j]->peaks()[path[k].second];
							peak1->add(peak2,match);
							os << "\t" << peak1->mass();
						}
					}
					os << std::endl;

					// second peakset
					os << spectra[j]->name();
					for (size_t k=0; k<path.size(); k++) {
						double match = dot_m(path[k].first, path[k].second);
						if (match > option.pairwise_cutoff()){
							Peak* peak1 = spectra[i]->peaks()[path[k].first];
							Peak* peak2 = spectra[j]->peaks()[path[k].second];
							peak2->add(peak1,match);
							os << "\t" << peak2->mass();
						}
					}
					os << std::endl;

					// score
					os << "Score:";
					for (size_t k=0; k<path.size(); k++)
						if (dot_m(path[k].first, path[k].second)>option.pairwise_cutoff())
							os << "\t" << dot_m(path[k].first, path[k].second);
					os << std::endl;

					// average
					os << "Average:";
					for (size_t k=0; k<path.size(); k++)
						if (dot_m(path[k].first, path[k].second)>option.pairwise_cutoff())
							os << "\t"
									<< (spectra[i]->peaks()[path[k].first]->mass()+
											spectra[j]->peaks()[path[k].second]->mass())/2;
					os << std::endl << std::endl;
				}
      }
    os.close();
  }

  void consensus(const std::vector<Spectrum* >& spectra,
								 const Parameter& option)
  {
    std::vector<Cluster> cluster(1);
		// avoid reallocations (we use pointers to elements in cluster)
		cluster.reserve(10000);
    for (size_t i=0;i<spectra.size(); i++){
      for (size_t j=0;j<spectra[i]->peaks().size(); j++){
				// assigning the peak, its matched peaks, and their matches,
				// i.e. recursively
				if (spectra[i]->peaks()[j]->assign(cluster)){
					// if peak was not already assigned
					Cluster c;
					cluster.push_back(c);
				}
      }
    }

    // sorting the clusters with respect to mass average
    sort(cluster.begin(),cluster.end());

    std::ofstream os(option.consensus_file().c_str());
    os << "Width: " << option.sigma() << "\n"
       << "Pairwise cutoff: " << option.pairwise_cutoff() << "\n"
       << "Consensus cutoff: " << option.consensus_cutoff() << "\n\n";
    cluster[0].write_header(os,spectra);
    for (size_t i=0; i<cluster.size(); i++){
      if (cluster[i].averager().n() > option.consensus_cutoff() )
				cluster[i].write(os, spectra);
    }
    os.close();
  }


  void multiple(const std::vector<Spectrum* >& spectra,const Parameter& option)
  {
    std::ofstream os(option.multiple_file().c_str());
    os << "Width: " << option.sigma() << "\n"
       << "Pairwise cutoff: " << option.pairwise_cutoff() << "\n"
       << "Multiple cutoff: " << option.multiple_cutoff() << "\n\n";
    for(size_t i=0; i<spectra.size(); i++){
      bool any_match = false;
      for(size_t j=0; j<spectra[i]->nof_peaks(); j++){
				double score =
					spectra.size()-1
					? spectra[i]->peaks()[j]->score_sum()/(spectra.size()-1) : 0;
				if (score>option.multiple_cutoff()){
					if (!any_match){
						any_match=true;
						os << spectra[i]->name();
					}
					os << "\t" << spectra[i]->peaks()[j]->mass();
				}
      }
      if (any_match){
				os << "\n";
				any_match = false;
				for(size_t j=0; j<spectra[i]->nof_peaks(); j++){
					double score =
						spectra.size()-1
						? spectra[i]->peaks()[j]->score_sum()/(spectra.size()-1) : 0;
					if (score>option.multiple_cutoff()){
						if (!any_match){
							any_match=true;
						}
						os << "\t" << score;
					}
				}
				os << "\n";
      }
    }
    os.close();
  }


	void read_id(std::vector<std::string>& id, const std::string& str)
	{
		id.resize(0);
		std::ifstream is(str.c_str());
		std::string s;
		for (int i = 0; getline(is, s, '\n'); i++)
			id.push_back(s);
	}


	void read_vector(std::vector<double>& v,std::string& str)
	{
		std::istringstream s(str);
		while (!s.eof()) {
			double d;
			s >> d;
			if (!s.fail())
				v.push_back(d);
		}
	}



	std::istream& operator>>(std::istream& s,
													 std::vector<std::vector<double> >& m)
	{
		while (!s.eof()) {
			std::string row;
			char ch;
			while (((ch=s.get())!='\n') && s.good())
				row+=ch;
			if (row.size() || s.good()) {	// if stream good, read an empty peak set
				std::vector<double> v;
				read_vector(v,row);
				m.push_back(v);
			}
		}
		return s;
	}



	std::ostream& operator<<(std::ostream& s,
													 const std::vector<std::vector<double> >& m)
	{
		using namespace std;
		for (vector<vector<double> >::const_iterator i=m.begin(); i!=m.end(); i++) {
			for (vector<double>::const_iterator j=i->begin(); j!=i->end(); j++)
				s << *j << ' ';
			s << '\n';
		}
		return s;
	}
	
}} // of namespace speclust and namespace theplu
