// $Id: psd.cc 228 2012-09-13 22:45:17Z peter $

// Copyright (C) 2004, 2005 Peter Johansson
// Copyright (C) 2006 Markus Ringnér
// Copyright (C) 2011 Peter Johansson
//
// This file is part of speclust, http://bioinfo.thep.lu.se/speclust.html
//
// Speclust is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 3 of the License, or (at your option) any later
// version.
//
// Speclust is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// speclust. If not, see <http://www.gnu.org/licenses/>.

#include "Distance.h"
#include "DistanceConservative.h"
#include "DistanceEuclid.h"
#include "DistanceLiberal.h"
#include "DistancePearson.h"

#include <yat/utility/CommandLine.h>
#include <yat/utility/Matrix.h>
#include <yat/utility/OptionArg.h>
#include <yat/utility/OptionHelp.h>
#include <yat/utility/utility.h>

#include <cassert>
#include <fstream>
#include <iostream>
#include <iterator>
#include <sstream>
#include <vector>

void calculate_distance(theplu::yat::utility::Matrix& d,
												const std::vector<std::vector<double> >& ps,
												double sigma,
												const theplu::speclust::Distance& metric)
{
	for (unsigned int i=0; i<ps.size()-1; i++)
		for (unsigned int j=i+1; j<ps.size(); j++)
			d(i,j) = d(j,i) = metric(ps[i], ps[j], sigma);
}


int main(int argc, char* argv[])
{
	using namespace std;

	theplu::yat::utility::CommandLine cmd;
	theplu::yat::utility::OptionHelp help(cmd);

	theplu::yat::utility::OptionArg<std::string>
		cmd_metric(cmd, "m,metric",
				 "valid metrics are: euclid, pearson, conserv, or liberal [pearson]");
	cmd_metric.value("pearson");
	cmd_metric.print_arg("=METRIC");

	theplu::yat::utility::OptionArg<double>
		cmd_sigma(cmd, "s,sigma", "width used in peak match score [1.0]");
	cmd_sigma.value(1.0);
	cmd_sigma.print_arg("=FLT");

	std::ostringstream ss;
	ss << "Calculate the distance between peak sets.\n\n";
	ss << "Program reads peaks sets from stdin in format as from psm\n"
		 << "and sends resulting distance matrix to stdout.\n";
	help.synopsis() = ss.str();

	try {
		cmd.parse(argc, argv);
	}
	catch (theplu::yat::utility::cmd_error& e) {
		std::cerr << e.what() << std::endl;
		return EXIT_FAILURE;
	}

	theplu::speclust::Distance* metric=NULL;
	if (cmd_metric.value() == "pearson")
 		metric = new theplu::speclust::DistancePearson;
	else if (cmd_metric.value() == "euclid")
 		metric = new theplu::speclust::DistanceEuclid;
	else if (cmd_metric.value() == "liberal")
		metric = new theplu::speclust::DistanceLiberal;
	else if (cmd_metric.value() == "conserv")
 		metric = new theplu::speclust::DistanceConservative;
	else {
		std::cerr << cmd.program_name() << ": invalid value for option --metric: "
							<< cmd_metric.value() << "\n";
		return EXIT_FAILURE;
	}

	vector<vector<double> > peaksets;
	theplu::yat::utility::load(std::cin, peaksets, '\0', '\n', true, false);

	theplu::yat::utility::Matrix distance(peaksets.size(), peaksets.size());
	assert(metric);
	calculate_distance(distance, peaksets, cmd_sigma.value(), *metric);
	delete metric;
	cout << distance << endl;

	return EXIT_SUCCESS;
}
