// $Id: DrawTree.cc 281 2013-02-28 14:53:03Z markus $

// Copyright (C) 2002 Carl Troein
// Copyright (C) 2004 Markus Ringnér
// Copyright (C) 2005, 2006 Peter Johansson
// Copyright (C) 2007 Jari Häkkinen
// Copyright (C) 2011 Peter Johansson
//
// This file is part of speclust, http://bioinfo.thep.lu.se/speclust.html
//
// Speclust is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 3 of the License, or (at your option) any later
// version.
//
// Speclust is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// speclust. If not, see <http://www.gnu.org/licenses/>.

// Speclust
#include "DrawTree.h"

// Additional packages
#include <cgraph.h>

// Standard C++
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <fstream>
#include <iostream>
#include <sstream>

#include <unistd.h>

namespace theplu {
namespace speclust {

  DrawTree::DrawTree(const vector<TreeNode>& nodes,
										 const vector<string>& leaves)
    : maxpoints_(nodes.size()+1), leaves_(leaves),
      ps_nodewidth_(12)
  {
    // Calculate the height of the picture
    unsigned int maxleaf=0;
    for(unsigned int i=0; i<leaves_.size(); i++){
      if(maxleaf < leaves_[i].size())
				maxleaf = leaves_[i].size();
    }
		ps_height_=maxleaf*(ps_nodewidth_+1)/2+100;


    // Calculate the order the leafs will be printed in.
    leaf_order(nodes, nodes.size()-1);

    // convert nodes to Cecilias format
    for(unsigned int i=0;i<nodes.size()+1;i++)
      nodes_.push_back(TreeNode(-1,-1,0));
    for(unsigned int i=0;i<nodes.size();i++) {
      int a=nodes[i].i();
      int b=nodes[i].j();
      if(a<0)
				a=nodes.size()-a;
      if(b<0)
				b=nodes.size()-b;
      nodes_.push_back(TreeNode(a,b,nodes[i].distance()));
    }

    ofstream os("sorted.list");
    for(size_t i=0;i<leaf_order_.size();i++) {
			assert(i<leaf_order_.size());
			assert(static_cast<size_t>(leaf_order_[i]) < leaves_.size());
      os << leaves_[leaf_order_[i]] << "\n";
		}
    os.close();
  }


  void DrawTree::leaf_order(const vector<TreeNode>& nodes, int node)
  {
    if(nodes[node].i()<0)
      leaf_order(nodes, abs(nodes[node].i())-1);
    else {
			assert(node<static_cast<int>(nodes.size()));
      leaf_order_.push_back(nodes[node].i());
    }
    if(nodes[node].j()<0)
      leaf_order(nodes, abs(nodes[node].j())-1);
    else {
			assert(node<static_cast<int>(nodes.size()));
      leaf_order_.push_back(nodes[node].j());
    }
  }


  void DrawTree::make_picture(string filename) const
  {
    double maxdist = 0;
    for(unsigned int i=0; i<nodes_.size(); i++){
      if(maxdist < nodes_[i].distance())
				maxdist = nodes_[i].distance();
    }
    if(maxdist==0)
      maxdist = 1;

    //gr bild (maxpoints x 12 + annotation) x ps_height_
    char bbox[64];
		sprintf(bbox, "0 %d %d %d", 0, maxpoints_ * ps_nodewidth_ + 40, ps_height_);
    cg_setboundingbox(bbox);
    unlink(filename.c_str());
    cg_set_output_filename( const_cast<char*>(filename.c_str()) );
    cg_init(1, 0, 1.0/72.0);
    cg_aorigin(0.0, 0.0);
    cg_linewidth(1.0);
    cg_font(10, 30);


    double palette[13*3] =
      {0, 0, 0,//black
       1, 0, 0,//red
       0, 0.5, 0,//green
       0, 0, 0.8,//blue
       0.9, 0, 0.9,//cerise
       1, 0.67, 0,//orange
       0, 0.7, 0.6,//cyan,green
       1, 0.33, 0,//red orange
       0, 0.67, 1,//aqua,blue
       0.7, 0, 0.4,//dark red
       0.55, 0, 1,//blue-purple
       0.6, 0.3, 0,//brown
       0.5, 0.5, 0.5};//grey

    int p = 0;
    double xroot,yroot;
    draw_tree(maxdist, nodes_.size()-1, p, xroot, yroot,
							palette);

    // draw scale
    draw_scale(yroot);

    //cg_launch_preview(0);
    //cg_showpage();
  }

  void DrawTree::set_color(double* palette, int value) const
  {
    double r,g,b;
    if(value!=0)
      value = 3*((value-1)%12)+3;
    r=palette[value];
    g=palette[value+1];
    b=palette[value+2];
    cg_rgbcolor(r,g,b);
  }

  void DrawTree::draw_scale(double yroot) const
  {
    cg_linewidth(0.5);
    cg_font(8, 30);
    double ticksize=ps_nodewidth_/3;
    cg_move((maxpoints_ + 1.0) * ps_nodewidth_, ps_height_-100);
    cg_line((maxpoints_ + 1.0) * ps_nodewidth_,yroot);


    std::stringstream ss;
    ss.precision(3);

    double ticks_dist=0.25;
    if(nodes_[nodes_.size()-1].distance()<0.8)
      ticks_dist=0.20;
    if(nodes_[nodes_.size()-1].distance()<0.40)
      ticks_dist=0.05;
    if(nodes_[nodes_.size()-1].distance()<0.16)
      ticks_dist=0.02;
    if(nodes_[nodes_.size()-1].distance()<0.08)
      ticks_dist=0.01;

    double y=ps_height_-100;
    double value=0;
    while(value <= nodes_[nodes_.size()-1].distance()){
      ss.str("");
      ss << value;
      string name = ss.str();
      char buf[FILENAME_MAX];
      sprintf(buf, "%s", name.c_str());
      cg_textalign((maxpoints_ + 1.5) * ps_nodewidth_, y,
		   0, buf, 0, 1, 0 );	
      cg_move((maxpoints_ + 1.0) * ps_nodewidth_-ticksize/2, y);
      cg_line((maxpoints_ + 1.0) * ps_nodewidth_+ticksize/2,y);
      y+=(yroot-(ps_height_-100))*ticks_dist/nodes_[nodes_.size()-1].distance();
      value+=ticks_dist;
    }

    cg_stroke();
}

  void DrawTree::draw_tree(double maxdist, unsigned int node_no,
			    int &pointsabove, double &xcoord, double &ycoord,
			    double* palette) const
  {
    const TreeNode& node = nodes_[node_no];

    if(node.i()<0){ // write leaf names
      xcoord = (pointsabove + 0.5) * ps_nodewidth_;
      ycoord = ps_height_-100;
      std::stringstream ss;
      ss << leaves_[leaf_order_[pointsabove]];
      string name = ss.str();
      char buf[FILENAME_MAX];
      sprintf(buf, "%s", name.c_str());
      set_color(palette,0);
      cg_textalign(xcoord, ps_height_-105, 90.0, buf, 2, 1, 0 );
      pointsabove++;
    }
    else {
			double xi, yi, xj, yj;
      draw_tree(maxdist, node.i(), pointsabove, xi, yi,
								palette);
      draw_tree(maxdist, node.j(), pointsabove, xj, yj,
								palette);

      if(node.distance() >= 0 && yi < 1e10 && yj < 1e10) {
				xcoord = (xi+xj)*0.5;
				ycoord = ps_height_-100 + 95 * node.distance()/maxdist;
				if(yi > ycoord || yj > ycoord) {
					ycoord = max(yi,yj);
				}
				cg_move(xi,yi);
				cg_line(xi,ycoord);
				cg_line(xj,ycoord);
				cg_line(xj,yj);
				
				cg_stroke();
      }
      else {
				ycoord = 2e10;
      }
    }
  }

}} // end of namespaces
